# -*- coding: utf-8 -*-

from freezegun import freeze_time
from odoo import Command
from odoo.tests import tagged, HttpCase
from odoo.addons.account_reports.tests.common import TestAccountReportsCommon


@tagged("post_install", "-at_install")
class TestDeferredReports(TestAccountReportsCommon, HttpCase):
    @classmethod
    def setUpClass(cls, chart_template_ref=None):
        super().setUpClass(chart_template_ref=chart_template_ref)
        cls.deferred_expense_report = cls.env.ref("account_reports.deferred_expense_report")
        cls.deferred_revenue_report = cls.env.ref("account_reports.deferred_revenue_report")
        cls.handler = cls.env["account.deferred.expense.report.handler"]

        cls.expense_accounts = [cls.env["account.account"].create({
            "name": f"Expense {i}",
            "code": f"EXP{i}",
            "account_type": "expense",
        }) for i in range(3)]
        cls.revenue_accounts = [cls.env["account.account"].create({
            "name": f"Revenue {i}",
            "code": f"REV{i}",
            "account_type": "income",
        }) for i in range(3)]

        cls.company = cls.company_data["company"]
        cls.deferral_account = cls.company_data["default_account_deferred_expense"]
        cls.company.deferred_journal_id = cls.company_data["default_journal_misc"].id
        cls.company.deferred_expense_account_id = cls.company_data["default_account_deferred_expense"].id
        cls.company.deferred_revenue_account_id = cls.company_data["default_account_deferred_revenue"].id

        # Commission Plans, Grade and Referrers
        cls.ready_partner_plan = cls.env["commission.plan"].create({"name": "Ready Partner Plan"})
        cls.gold_partner_plan = cls.env["commission.plan"].create({"name": "Gold Partner Plan"})

        cls.ready_partner_grade = cls.env["res.partner.grade"].create({
            "name": "Ready Partner",
            "default_commission_plan_id": cls.ready_partner_plan.id,
        })
        cls.gold_partner_grade = cls.env["res.partner.grade"].create({
            "name": "Gold Partner",
            "default_commission_plan_id": cls.gold_partner_plan.id,
        })

        cls.ready_partner = cls.env["res.partner"].create({
            "name": "Ready Partner",
            "grade_id": cls.ready_partner_grade.id,
            "commission_plan_id": cls.ready_partner_plan.id,
            "country_id": cls.env.ref("base.fr").id,
        })
        cls.gold_partner = cls.env["res.partner"].create({
            "name": "Gold Partner",
            "grade_id": cls.gold_partner_grade.id,
            "commission_plan_id": cls.gold_partner_plan.id,
            "country_id": cls.env.ref("base.be").id,
        })
        cls.customer = cls.env["res.partner"].create({
            "name": "Customer",
            "country_id": cls.env.ref("base.be").id,
        })

    def create_invoice(self, invoice_lines, move_type="in_invoice", invoice_date="2023-01-01", post=True):
        journal = self.company_data["default_journal_sale"]
        if move_type.startswith("in_"):
            journal = self.company_data["default_journal_purchase"]
        move = self.env["account.move"].create({
            "move_type": move_type,
            "partner_id": self.partner_a.id,
            "date": invoice_date,
            "invoice_date": invoice_date,
            "journal_id": journal.id,
            "invoice_line_ids": [
                Command.create({
                    "product_id": self.product_a.id,
                    "quantity": 1,
                    "account_id": invoice_line[0].id,
                    "price_unit": invoice_line[1],
                    "deferred_start_date": invoice_line[2],
                    "deferred_end_date": invoice_line[3],
                    "hosting": invoice_line[4],
                }) for invoice_line in invoice_lines
            ]
        })
        if post:
            move.action_post()
        return move

    def create_sub(self, partner, template, start_date, pricelist=None, users=1, hosting=None, referrer=None, invoice_periods=None):
        users_prodcateg = self.env.ref("openerp_enterprise.product_category_odoo_users")
        with freeze_time(start_date):
            sub = self.env["sale.order"].create({
                "partner_id": partner.id,
                "sale_order_template_id": self.env.ref(template).id,
            })
            sub._onchange_sale_order_template_id()
            if pricelist:
                sub.pricelist_id = pricelist
            for line in sub.order_line:
                if line.product_id.categ_id == users_prodcateg:
                    line.product_uom_qty = users
            if hosting is not None:
                sub.hosting = hosting
            if referrer:
                sub.referrer_id = referrer
            sub.action_confirm()
            for _period in range(invoice_periods or 0):
                with freeze_time(sub.next_invoice_date):
                    sub._create_invoices()._post()
            return sub

    def get_options(self, from_date, to_date, report=None, default_options=None):
        report = report or self.deferred_expense_report
        return self._generate_options(report, from_date, to_date, default_options=default_options)

    def test_deferred_revenue_report_split_by_hosting(self):
        """
        Test the deferred revenue report with the "month" method.
        """
        self.company.deferred_amount_computation_method = "month"
        revenue_lines = [
            # 4 full months (=250/month)
            [self.revenue_accounts[0], 1000, "2023-01-01", "2023-04-30", "saas"],
            # 3 full months + 15 days (=300/month)
            [self.revenue_accounts[0], 1050, "2023-01-16", "2023-04-30", "paas"],
            # 3 full months + 15 days (=350/month)
            [self.revenue_accounts[1], 1225, "2023-01-01", "2023-04-15", "premise"],
            # 2 full months + 10 days + 14 days (=600/month)
            [self.revenue_accounts[2], 1680, "2023-01-21", "2023-04-14", "n/a"],
            # 15 days (=450/month)
            [self.revenue_accounts[2],  225, "2023-04-01", "2023-04-15",  False],
        ]
        self.create_invoice(revenue_lines, "out_invoice")

        # December 2022
        options = self.get_options(
            "2022-12-01", "2022-12-31", self.deferred_revenue_report, {"hosting_split": True}
        )
        lines = self.deferred_revenue_report._get_lines(options)
        self.assertLinesValues(
            lines,
            #    Name   Total   Not Started   Before   Current   Later
            [    0,     1,      2,            3,       4,        5     ],  # noqa: E201,E202
            [],
            options,
        )

        # January 2023
        options = self.get_options(
            "2023-01-01", "2023-01-31", self.deferred_revenue_report, {"hosting_split": True}
        )
        lines = self.deferred_revenue_report._get_lines(options)
        self.assertLinesValues(
            lines,
            #    Name                       Total          Not Started   Before  Current              Later
            [    0,                         1,             2,            3,      4,                   5                         ],  # noqa: E201,E202
            [
                ("REV0 Revenue 0",          1000 + 1050,   0,            0,      250 + 150,           750 + 900                 ),  # noqa: E202
                ("Hosting: Online",         1000,          0,            0,      250,                 750                       ),  # noqa: E202
                ("Hosting: Odoo.sh",        1050,          0,            0,      150,                 900                       ),  # noqa: E202
                ("Total REV0 Revenue 0",    1000 + 1050,   0,            0,      250 + 150,           750 + 900                 ),  # noqa: E202

                ("REV1 Revenue 1",          1225,          0,            0,      350,                 875                       ),  # noqa: E202
                ("Hosting: On Premise",     1225,          0,            0,      350,                 875                       ),  # noqa: E202
                ("Total REV1 Revenue 1",    1225,          0,            0,      350,                 875                       ),  # noqa: E202

                ("REV2 Revenue 2",          1680 + 225,    225,          0,      600 * (10 / 30) + 0, 600 * (2 + 14 / 30) + 225 ),  # noqa: E202
                ("Hosting: Not Applicable", 1680,          0,            0,      600 * (10 / 30),     600 * (2 + 14 / 30)       ),  # noqa: E202
                ("Hosting: (not set)",      225,           225,          0,      0,                   225                       ),  # noqa: E202
                ("Total REV2 Revenue 2",    1680 + 225,    225,          0,      600 * (10 / 30) + 0, 600 * (2 + 14 / 30) + 225 ),  # noqa: E202

                ("Total",                   5180,          225,          0,      950,                 4230                      ),  # noqa: E202
            ],
            options,
        )

    def test_deferred_revenue_report_split_by_commission_plan(self):
        """
        Test the deferred revenue report with the "month" method.
        """
        self.company.deferred_amount_computation_method = "month"
        yearly_template = "openerp_enterprise.custom_template_odoo_yearly"
        yearly_partnership_template = "openerp_enterprise.quotation_template_odoo_partnership"

        # Setup products for test
        plan_yearly = self.env.ref("openerp_enterprise.subscription_plan_odoo_yearly_plan")
        plan_yearly_partnership = self.env.ref("openerp_enterprise.subscription_plan_odoo_yearly_official_partnership")

        product_reccuring_prices = [
            ("openerp_enterprise.product_maintenance",               self.revenue_accounts[0], plan_yearly,             180.0),
            ("openerp_enterprise.product_odoo_user_custom",          self.revenue_accounts[0], plan_yearly,             600.0),
            ("openerp_enterprise.product_odoo_user_custom_discount", self.revenue_accounts[1], plan_yearly,            -120.0),
            ("openerp_enterprise.odoo_learning_platform_access",     self.revenue_accounts[2], plan_yearly_partnership, 1200.0),
        ]
        for product_xmlid, revenue_account, plan, price in product_reccuring_prices:
            product = self.env.ref(product_xmlid)
            product.property_account_income_id = revenue_account
            product.product_subscription_pricing_ids = [
                Command.clear(),
                Command.create({"plan_id": plan, "price": price})
            ]

        # Partnership: 1200 for 1 year (100 / month)
        self.create_sub(
            self.gold_partner, yearly_partnership_template, "2022-07-01",
            hosting="paas", invoice_periods=1,
        )
        self.gold_partner.write({
            "grade_id": self.gold_partner_grade.id,
            "commission_plan_id": self.gold_partner_plan.id,
        })
        # Customer (Online): (600 - 120) = 480 for 1 year for 1 user (50 - 10 = 40 / month)
        self.create_sub(
            self.customer, yearly_template, "2023-01-01",
            hosting="saas", invoice_periods=1, referrer=self.ready_partner,
        )
        # Customer (Premise): (600 - 120) = 480 for 1 year for 10 users = 4800 (500 - 100 = 400 / month)
        self.create_sub(
            self.customer, yearly_template, "2022-06-01",
            users=10, hosting="premise", invoice_periods=1, referrer=self.gold_partner
        )

        # January 2023
        options = self.get_options(
            "2023-01-01", "2023-01-31", self.deferred_revenue_report, {"commission_plan_split": True}
        )
        lines = self.deferred_revenue_report._get_lines(options)
        self.assertLinesValues(
            lines,
            #    Name                               Total          Not Started   Before     Current    Later
            [    0,                                 1,             2,            3,         4,         5          ],  # noqa: E201,E202
            [
                ("REV0 Revenue 0",                  6000 + 600,    0,            3500,      500 + 50,  2000 + 550 ),  # noqa: E202
                ("Comm. Plan: Gold Partner Plan",   6000,          0,            500 * 7,   500,       500 * 4    ),  # noqa: E202
                ("Comm. Plan: Ready Partner Plan",   600,          0,            0,         50,        50 * 11    ),  # noqa: E202
                ("Total REV0 Revenue 0",            6000 + 600,    0,            3500,      500 + 50,  2000 + 550 ),  # noqa: E202

                ("REV1 Revenue 1",                  -1320,         0,            -700,     -110,      -510        ),  # noqa: E202
                ("Comm. Plan: Gold Partner Plan",   -1200,         0,            -100 * 7, -100,      -100 * 4    ),  # noqa: E202
                ("Comm. Plan: Ready Partner Plan",  -120,          0,            0,        -10,       -110        ),  # noqa: E202
                ("Total REV1 Revenue 1",            -1320,         0,            -700,     -110,      -510        ),  # noqa: E202

                ("REV2 Revenue 2",                  1200,          0,            600,       100,       500        ),  # noqa: E202
                ("Comm. Plan: (not set)",           1200,          0,            600,       100,       500        ),  # noqa: E202
                ("Total REV2 Revenue 2",            1200,          0,            600,       100,       500        ),  # noqa: E202

                ("Total",                           6480,          0,            3400,      540,       2540       ),  # noqa: E202
            ],
            options,
        )
